
import 'package:path/path.dart';
import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:math';

import 'remotescheme.dart';

import 'filestate.dart';
import 'filestateprovider.dart';
import 'remotemethod.dart';

class RemoteSource {
  File _file;
  String _url;
  Uri _uri;
  int _totalSize = 0;
  int _currentSize = 0;
  bool _completed = false;
  RemoteScheme _scheme = RemoteScheme.HTTP;
  RemoteMethod _method = RemoteMethod.GET;
  Map<String, String> _body = new Map<String, String>();
  Map<String, String> _headers = new Map<String, String>();
  FileStateProvider sp;
  StreamController<Null> _download = new StreamController<Null>.broadcast();

  RemoteSource(String url) {
    this._url = url;
    Uri uriBefore = Uri.parse(this._url);
    if (uriBefore.scheme.toLowerCase() == 'http') {
      this._uri = Uri.http(
          uriBefore.authority, uriBefore.path, uriBefore.queryParameters);
    } else {
      this._scheme = RemoteScheme.HTTPS;
      this._uri = Uri.https(
          uriBefore.authority, uriBefore.path, uriBefore.queryParameters);
    }
    sp = new FileStateProvider();
  }
  int getCurrent() {
    return this._currentSize;
  }

  int getTotalSize() {
    return this._totalSize;
  }

  bool isCompleted() {
    return this._completed;
  }
  Future<FileSystemEntity> delete() async {
      return this._file.delete();
  }

  void setBody(Map<String, String> body) {
    this._body = body;
    if (this._scheme == RemoteScheme.HTTP) {
      this._uri = Uri.http(this._uri.authority, this._uri.path, this._body);
    } else {
      this._uri = Uri.https(this._uri.authority, this._uri.path, this._body);
    }
  }

  void addHeader(String property, String value) {
    if (property.toLowerCase().trim() ==
            HttpHeaders.acceptEncodingHeader.toLowerCase().trim() ||
        property.toLowerCase().trim() ==
            HttpHeaders.contentEncodingHeader.toLowerCase().trim()) {
      throw 'Cannot Set Accept-Encoding or Content-Encoding';
    }
    this._headers[property] = value;
  }

  void setMethod(RemoteMethod method) {
    this._method = method;
  }

  void setScheme(RemoteScheme protocol) {
    this._scheme = protocol;
    if (this._scheme == RemoteScheme.HTTP) {
      this._uri = Uri.http(
          this._uri.authority, this._uri.path, this._uri.queryParameters);
    } else {
      this._uri = Uri.https(
          this._uri.authority, this._uri.path, this._uri.queryParameters);
    }
  }

  Future<int> length() {
    Completer<int> c = new Completer<int>();
    if (this._totalSize > 0) {
      new Timer(new Duration(microseconds: 0), () {
        c.complete(this._totalSize);
      });
    } else {
      new HttpClient().headUrl(this._uri).then((HttpClientRequest request) {
        request.headers
            .set(HttpHeaders.acceptEncodingHeader, 'accept-encoding=identity');
        this._headers.forEach((String property, String value) {
          request.headers.set(property, value);
        });
        return request.close();
      }).then((HttpClientResponse response) {
        try {
          this._totalSize = int.parse(response.headers['content-length'][0]);
          c.complete(this._totalSize);
        } catch (e) {
          this._totalSize = 0;
          c.complete(this._totalSize);
        }
      }).catchError((error) {});
    }
    return c.future;
  }

  File getFile() {
    return this._file;
  }

  Stream<Null> get downloaded {
    return _download.stream;
  }

  void download() async {
    Uri uri = Uri.parse(this._url);
    String pa = uri.path;
    if (pa.isEmpty) {
      var random = new Random();
      pa = "file_" + random.nextInt(10000).toString();
    }
    String base = basename(uri.path);
    new File(Directory.systemTemp.path +
            '/' +
            (base.contains('.') ? base : base + '.json'))
        .create()
        .then((File f) {
      this._file = f;
      FileMode fm = FileMode.write;
      int i = 0;
      Future<HttpClientRequest> req;
      if (this._method == RemoteMethod.GET) {
        req = new HttpClient().getUrl(this._uri);
      } else {
        req = new HttpClient().postUrl(this._uri);
      }
      req.then((HttpClientRequest request) {
        this._headers.forEach((String property, String value) {
          request.headers.set(property, value);
        });
        return request.close();
      }).then((HttpClientResponse response) {
        response.transform(utf8.decoder).listen((contents) {
          if (i > 0) {
            fm = FileMode.append;
          }
          f.writeAsStringSync(contents, mode: fm);
          this._currentSize += contents.length;
          new Timer(new Duration(microseconds: 1), () {
            this.sp.pushState(this._currentSize, FileStateProvider.WAITING);
          });
          i++;
        }, onDone: () {
          new Timer(new Duration(microseconds: 1), () {
            this._file.length().then((int l) {
              this._completed = true;
              this._totalSize = l;
              this._currentSize = l;
              this.sp.pushState(l, FileStateProvider.COMPLETED);
              new Timer(new Duration(microseconds: 10), () {
                sp.close();
              });
              _download.add(null);
              _download.close();
            });
          });
          //sink.close();
          i = null;
        }, onError: (error) {
          _download.addError(error);
          _download.close();
        });
      });
    }).catchError((error) {
      _download.addError(error);
      _download.close();
    });
  }

  Stream<List<int>> openRead(int start, [int end = 0]) {
    StreamController<List<int>> cc = new StreamController<List<int>>();
    if (end == 0 && this._completed) {
      this._file.openRead(start).listen((List<int> line) {
        cc.add(line);
      }, onDone: () {
        cc.close();
      }, onError: (e) {
        cc.addError(e);
      });
    } else if (end > 0 && end <= this._currentSize) {
      this._file.openRead(start, end).listen((List<int> line) {
        cc.add(line);
      }, onDone: () {
        cc.close();
      }, onError: (e) {
        cc.addError(e);
      });
    } else {
      if (end == 0) {
        this.sp.pushState(this._totalSize, FileStateProvider.CALLED);
      } else {
        this.sp.pushState(end, FileStateProvider.CALLED);
      }
      StreamSubscription<FileState> fstream;
      fstream = sp.ready().listen((FileState data) {
        if (end == 0) {
          if (this._completed) {
            this._file.openRead(start).listen((List<int> line) {
              if (!cc.isClosed) {
                cc.add(line);
              } else {
                fstream.cancel();
              }
            }, onDone: () {
              cc.close();
            }, onError: (e) {
              cc.addError(e);
            });
          }
        } else if (data.length >= end) {
          this._file.openRead(start, end).listen((List<int> line) {
            if (!cc.isClosed) {
              cc.add(line);
            } else {
              fstream.cancel();
            }
          }, onDone: () {
            cc.close();
          }, onError: (e) {
            cc.addError(e);
          });
        }
      });
    }
    return cc.stream;
  }
}
